unit ICDocument;

(*	This file is part of the Internet Configuration system and is placed in the public domain for the benefit of all.

	This module implements the core of the IC document architecture, with the usual array of document
	open, close, and save commands. Because the only substance behind IC documents is an Internet
	Config preferences file, a document has a one-to-one relationship to an ICInstance that is the
	preference file associated with it. Because we can only open one document at a time, it's
	easy to keep track of these instances.

	When the user closes a preference panel, the application saves the preferences for each item in that
	window back to the backing document (using the WhatFlush routines). Of course, at the user
	level closing a window is not the same as saving your changes, so we have to make sure that
	these changes aren't saved back to the real IC preferences file. The way we do this is by
	creating a gScratchFile that's a copy of the file we're editing, and doing all our temporary
	saves to that file. Then, when the user hits save, we simply overwrite the original file with our
	scratch file.
	
	Handling the scratch file is quite tricky. When we're asked to open a document,
	we create a scratch file and duplicate the document to that file. We then open
	that file directly using ICSpecifyConfigFile. All reads and writes are then
	vectored through the gInstance. When the user hits save, we simply
	overwrite the existing file with the scratch file en masse. When the user hits
	close, we delete the scratch file.
	
	One thing we do is make sure that we record the seed (see the IC Programming
	Docs for a description of seeds) for the document when we open it. We then
	check the seed before overwriting that document with the scratch file. So if
	some other application has been changing the main Internet Preferences file
	while we've been working on our copy, we can at least warn the user.
	
	The IC application can intimate knowledge of the internals of the IC component.
	All other applications are required to use the IC API to access the the
	IC preferences for all purposes.  The IC application uses the IC API for
	most purposes, but it also:
	
	1) knows where IC places the default preferences file,
	2) knows that all IC information is contained in this file, so that copying that
		file is enough to create a duplicate of the IC information,
	3) uses the ICSpecifyConfigFile to target the duplicate directly.
	4) knows how to install and register the IC Extension
	5) knows that it can open the IC Extension to get default preferences
	
	It's a long term goal of IC to remove this specialist knowledge about
	the component from the application. Currently I don't have a plan for
	doing this, but I hope to do it eventually.
*)

interface

	uses
		Files,
		Windows,
		ICTypes;

	(* ***** File Menu Commands ***** *)
	
	function DoOpenDefaultDocument: OSStatus;
		(* This routine is called when the application starts up (or in response to an Open
			Application AppleEvents) and it opens the default Internet Preferences file. It's
			distinct from DoNewDocument because that routine creates a new, untitled, preferences 
			document, whereas this routine opens the default one, creating it if it's not there. 
			This is one of the places where the IC human interface differs from a standard Macintosh
			application.
			
			This routine is also used by the Open Internet Preferences menu command.
		*)

	function DoNewDocument: OSStatus;
		(* This routine implements the standard behaviour of the New item on the File menu. *)
	
	function DoOpenDocumentUsingStandardFile: OSStatus;
		(* This routine implements the standard behaviour of the Open item on the File menu. 
			It basically calls Standard File to find a file and then DoOpenDocument to open it.
		*)
	
	function DoOpenDocument (fss: FSSpec): OSStatus;
		(* Open an existing preferences document. *)
	
	function DoCloseDocumentWindow (wp: WindowPtr): OSStatus;
		(* Closes a document window. This is distinct from DoCloseDocument because this routine
			closes just one window, whereas DoCloseDocument closes the document, which closes all the
			windows and the document to boot. This routine detects if you're closing the main
			document window, and calls DoCloseDocument if you are. It also allows you to close
			the About Box window.
		*)

	function DoCloseDocument: OSStatus;
		(* Closes the current preferences document. Returns userCanceledErr if the user cancels the operation. *)
	
	function DoSaveDocument: OSStatus;
		(* This routine implements the standard behaviour of the Save item on the File menu. *)
	
	function DoSaveAsDocument: OSStatus;
		(* This routine implements the standard behaviour of the Save As item on the File menu. *)

	function DoQuit : OSStatus;
		(* Quits the application. Because we can only have one document open at a time, this is a very 
			wrapper around DoCloseDocument. It exists because there are multiple ways to quit (ie from
			AppleEvents and from the menus), and it's worthwhile extracting the common code here.
			This is in ICDocument because the menu handling code is in ICEvents and it needs access
			to this routine.
		*)

	(* ***** Getters and Setters ***** *)
	
	function GetDocumentInstance: ICInstance;
		(* Gets the ICInstance associated with the current document.  Note that
			the ICInstance is created and destroyed when the application is opened
			and quit, not when the document is opened and closed. When the document
			is closed, the gInstance is left pointing to the document until a new
			document is opened. Do not call GetDocumentInstance when there is no document
			open.
		*)
	
	function IsDocumentOpen: Boolean;
		(* Returns true if there is a document open. *)
	
	function IsDocumentDirty: Boolean;
		(* Returns true if the document is dirty, ie needs to be saved. *)
	
	function IsDocumentNew: Boolean;
		(* Returns true if the document has just been created. IC uses this to enable the
			Save command even though the document isn't dirty.
		*)

	function IsDocumentLocked: Boolean;
		(* Returns true if the current document is locked. *)

	procedure DirtyDocument;
		(* This routine dirtys the current document. *)
	
	function GetDocumentName: Str255;
		(* Gets the name of the current document. Used to set the first item of the Windows menu
			and other things internal to this module.
		*)
	
	procedure GetDocumentCurrentFile(var currentFile : FSSpec);
		(* Gets the FSSpec associated with the currently open document.
			This information is only valid if there is a document open,
			ie IsDocumentOpen has returned true.
		*)

	(* ***** Misc Stuff ***** *)
	
	function FindInternetPreferences (var defaultConfig: FSSpec): OSStatus;
		(* This routine attempts to find the default Internet Preferences file,
			ie the one in the preference folder, following aliases if necessary.
			Returns fnfErr if the file does not exist yet and in that case sets
			defaultConfig to point to the place where the file should be created.
			This is one of the places where the IC application can intimate 
			knowledge of the internals of the IC component.
		*)
	
	function CreateSpecifiedInstance(var inst : ICInstance) : OSStatus;
	procedure DestroySpecifiedInstance(inst : ICInstance);
		(* These two utility functions are designed for parts of the application
			that require access to IC preferences but aren't guaranteed to
			have a document open at the time, eg the GURL event
			handler.  CreateSpecifiedInstance either
			returns gInstance (if IsDocumentOpen is true) or a
			newly created ICInstance.  DestroySpecifiedInstance
			disposes of the created instance (if it isn't gInstance
			of course).  If CreateSpecifiedInstance returns an
			error, something really wacky has happened and you should not call
			DestroySpecifiedInstance.
		*)

	function InitICDocument: OSStatus;
		(* Initialises this module. *)

	procedure TermICDocument;
		(* Shuts down this module. *)

implementation

	uses
		Folders, 
		Resources, 
		Dialogs, 
		Aliases, 
		ToolUtils,

		ICTypes, 
		ICAPI, 
		ICKeys, 

		ICResources, 
		ICGlobals, 
		ICStringLists, 
		ICMiscSubs, 
		ICDialogs, 
		ICWindowGlobals, 
		ICWindows, 
		ICInstall,
		ICStandardFile;

	var
		gInstance: ICInstance;
		(* gInstance is our primary connection to the Internet Config component.
			All of our preference reads and writes are vectored through the component,
			and hence through this instance.
		*)

		gCurrentFile: FSSpec;
		(* This is the FSSpec for the document we're currently editing.

			We also overload the name field. If it's empty, it means that the
			current document is a new untitled document.
		*)
				
		gSeedForCurrentFile: longint;
		(* This is the seed for the file we're currently editing. *)
		
		gScratchFile: FSSpec;			(* open if name <> '' *)
		(* This is the FSSpec for the scratch file. When a document is open,
			the ICInstance is pointing at this file, ready to support reading and writing
			of preference data.
			
			We also overload the name field. If it's empty, it means we have
			no document open.
		*)

		(* The following Booleans simply track the values described by
			their accessor functions in the interface part.
		*)
		gDirtyDocument: Boolean;
		gLockedDocument : Boolean;

	(* ***** Getters and Setters ***** *)
	
	function GetDocumentInstance: ICInstance;
	begin
		GetDocumentInstance := gInstance;
	end; (* GetDocumentInstance *)

	function IsDocumentOpen: Boolean;
	begin
		IsDocumentOpen := (gScratchFile.name <> '');
	end; (* IsDocumentOpen *)

	function IsDocumentDirty: Boolean;
	begin
		IsDocumentDirty := IsDocumentOpen and gDirtyDocument;
	end; (* IsDocumentDirty *)

	function IsDocumentNew: Boolean;
	begin
		IsDocumentNew := IsDocumentOpen and (gCurrentFile.name = '');
	end; (* IsDocumentNew *)

	function IsDocumentLocked: Boolean;
	begin
		IsDocumentLocked := IsDocumentOpen and gLockedDocument;
	end; (* IsDocumentLocked *)

	procedure DirtyDocument;
	begin
		gDirtyDocument := true;
	end; (* DirtyDocument *)

	function GetDocumentName: Str255;
		(* Returns the user visible name of the current document. We special
			case when the document name is '' (which is used to signify a new
			untitled document) and return 'untitled'.
		*)
	begin
		if gCurrentFile.name = '' then begin
			GetDocumentName := GetIndStr(rMiscStrings, strUntitled);
		end else begin
			GetDocumentName := gCurrentFile.name;
		end; (* if *)
	end; (* GetDocumentName *)
	
	procedure GetDocumentCurrentFile(var currentFile : FSSpec);
	begin
		currentFile := gCurrentFile;
	end; (* GetDocumentCurrentFile *)
	
	(* ***** Seed Management ***** *)

	procedure CurrentSeed (var seed: longint);
		(* Sets seed to be the current seed for gCurrentFile.
			The current implementation temporarily points gInstance to
			gCurrentFile, gets the seed and then points it back to gScratchFile.
			This is most probably not the best way of doing this.
			
			A seed value of 0 means that there is no gCurrentFile, ie
			the document is untitled.
		*)
		var
			err: OSStatus;
	begin
		seed := 0;
		if gCurrentFile.name <> '' then begin
			err := ICSpecifyConfigFile(gInstance, gCurrentFile);
			if err = noErr then begin
				err := ICGetSeed(gInstance, seed);
			end; (* if *)
			err := ICSpecifyConfigFile(gInstance, gScratchFile);
		end; (* if *)
	end; (* CurrentSeed *)

	function SameSeed (seed1, seed2: longint): Boolean;
		(* Returns true if seed1 matches seed2. This is used by DoSaveDocument
			to check whether we should warn the user that the seed for
			gCurrentFile has changed since we copied it into gScratchFile,
			ie someone has been modifying the Internet Preferences file
			while we were working on our scratch copy.
		*)
	begin
		SameSeed := (seed1 = seed2) or (seed1 = 0) or (seed2 = 0);
	end; (* SameSeed *)

	(* ***** Utility Routines for New Documents ***** *)
	

	function CreateTemporaryFile(vRefNum : integer; dirID : longint; var createdFile : FSSpec) : OSStatus;
		(* Create a temporary file in the directory given by vRefNum and dirID.
			Sets createdFile to the file spec of the created file.  Because
			HCreate is atomic, we're guaranteed to either own the resulting file or
			get a dupFNErr. If we get a duplicate error, we try again.
		*)
		const
			kTempFileName = 'Internet Config Temp ';
		var
			err : OSStatus;
			createRetryCounter : integer;
	begin
		createdFile.vRefNum := vRefNum;
		createdFile.parID := dirID;
		createRetryCounter := 0;
		repeat
			createdFile.name := concat(kTempFileName, DecStr(createRetryCounter));
			err := HCreate(createdFile.vRefNum, createdFile.parID, createdFile.name, 
											ICcreator, ICfiletype);
			createRetryCounter := createRetryCounter + 1;
		until err <> dupFNErr;
		CreateTemporaryFile := err;
	end; (* CreateTemporaryFile *)
	
	function CreateScratchFile (protoFile: FSSpecPtr): OSStatus;
		(* Create a scratch file based on the protoFile. The scratch file
			is created in the Temporary Items folder with a unique name.
			The FSSpec of the file is placed in gScratchFile.
			
			The protoFile is used to initialise the scratch file.
			If there is a protoFile, it is copied in order to create the scratch file.
			If there is no protoFile, ie protoFile is nil, the scratch file is created
			with the default preferences. 

			This is one of the places where the IC application can intimate 
			knowledge of the internals of the IC component.
		*)
		var
			err: OSStatus;
			junk : OSStatus;
			fileCreated : Boolean;
	begin
		fileCreated := false;
		
		(* First attempt to create a scratch file in the Temporary Items folder. *)
		err := FindFolder(kOnSystemDisk, kTemporaryFolderType, kCreateFolder, gScratchFile.vRefNum, gScratchFile.parID);
		if err = noErr then begin
			err := CreateTemporaryFile(gScratchFile.vRefNum, gScratchFile.parID, gScratchFile);
			fileCreated := (err = noErr);
		end; (* if *)

		(* Now initialise the scratch file based on protoFile. *)
		if err = noErr then begin
			if protoFile = nil then begin
				HCreateResFile(gScratchFile.vRefNum, gScratchFile.parID, gScratchFile.name);
				(* Temporary workaround bug in the IC component, which wasn't creating 
					a resource file properly. I suspect that this bug has long since been fixed,
					but there's little harm in leaving this code in.
				*)
			end else begin
				(* If we're using a protoFile (ie we're opening an existing file), 
					target IC at the file and do an ICBegin/ICEnd pair.  This forces
					IC to check the validity of the new file.  It's important that
					we do this before we clone the file into the scratch file
					because:
						a) IC will then give the correct filename in the error
							dialog, ie Internet Preferences rather than
							Internet Config Temp 0.
						b) the repairs that IC does to the file (if any) will
							be done to the primary file, not the scratch file.
							So the repairs will take effect even if the user
							doesn't choose save.
					-- Quinn, IC 1.4
				*)
				if err = noErr then begin
					err := ICSpecifyConfigFile(gInstance, protoFile^);
				end; (* if *)
				if err = noErr then begin
					err := ICBegin(gInstance, icReadWritePerm);
				end; (* if *)
				if err = noErr then begin
					err := ICEnd(gInstance);
				end; (* if *)
				
				if err = noErr then begin
					err := CopyFile(protoFile^, gScratchFile);
				end; (* if *)
				if err = noErr then begin
					(* Unlock the scratch file. CopyFile copies the file attributes, including the
						file lock. If we open a locked file, we still want the scratch file to be
						writable (and it should be, because it's in the Temporary Items folder),
						so we have to unlock it.
					*)
					err := HRstFLock(gScratchFile.vRefNum, gScratchFile.parID, gScratchFile.name);
				end; (* if *)
			end; (* if *)
		end; (* if *)
		
		(* Now target IC at the scratch file. *)
		if err = noErr then begin
			err := ICSpecifyConfigFile(gInstance, gScratchFile);
		end; (* if *)
		
		(* Clean up.  If we fail, we must set gScratchFile.name to empty to let
			IsDocumentOpen know that there's no document open.
		*)
		if (err <> noErr) and fileCreated then begin
			junk := HDelete(gScratchFile.vRefNum, gScratchFile.parID, gScratchFile.name);
		end; (* if *)
		if err <> noErr then begin
			gScratchFile.name := '';
		end; (* if *)
		
		CreateScratchFile := err;
	end; (* CreateScratchFile *)

	function InternalDoSave (fss: FSSpec): OSStatus;
		forward;
	
	(* ***** Newing and Opening ***** *)

	function FindInternetPreferences (var defaultConfig: FSSpec): OSStatus;
		(* See comment in interface part. *)
		var
			err: OSStatus;
			junkIsFolder, junkWasAlias: Boolean;
			junkFInfo : FInfo;
	begin
		(* We call FindFolder directly, presuming that the language glue will
			handle the pre-System 7 case.
		*)
		err := FindFolder(kOnSystemDisk, kPreferencesFolderType, kCreateFolder,
											defaultConfig.vRefNum, defaultConfig.parID);
		if err = noErr then begin
			err := ICDefaultFileName(gInstance, defaultConfig.name);
		end; (* if *)
		if err = noErr then begin
			if gHasAliasManager then begin
				(* Resolve a possible alias.  Note that this will trigger a fnfErr
					if the file is missing.  The caller expects this.
				*)
				err := ResolveAliasFile(defaultConfig, true, junkIsFolder, junkWasAlias);
			end else begin
				(* Without the Alias Manager, simply force a failure if the file doesn't exist. *)
				err := HGetFInfo(defaultConfig.vRefNum, defaultConfig.parID, defaultConfig.name, junkFInfo);
			end; (* if *)
		end; (* if *)
		FindInternetPreferences := err;
	end; (* FindInternetPreferences *)

	function DoOpenDefaultDocument: OSStatus;
		(* See comment in interface part. *)
		var
			defaultConfigFile: FSSpec;
			err: OSStatus;
	begin
		err := FindInternetPreferences(defaultConfigFile);
		if err = noErr then begin
			err := DoOpenDocument(defaultConfigFile);
		end; (* if *)
		if err = fnfErr then begin
			err := DoNewDocument;
			if err = noErr then begin
				err := InternalDoSave(defaultConfigFile);
			end; (* if *)
		end; (* if *)
		DoOpenDefaultDocument := err;
	end; (* DoOpenDefaultDocument *)

	function DoNewDocument: OSStatus;
		(* See comment in interface part. *)

		function AddDefaultPrefs : OSStatus;
			var
				err : OSStatus;
				junk : OSStatus;
				fileCreated : Boolean;
				extensionFileSpec : FSSpec;
				extensionRefNum : integer;
				prefCount : integer;
				prefResourceIndex : integer;
				prefData : Handle;
				junkID : integer;
				junkType : OSType;
				prefName : Str255;
		begin
			fileCreated := false;
			err := FindFolder(kOnSystemDisk, kTemporaryFolderType, kCreateFolder, 
													extensionFileSpec.vRefNum, extensionFileSpec.parID);
			if err = noErr then begin
				err := CreateTemporaryFile(extensionFileSpec.vRefNum, extensionFileSpec.parID, 
															extensionFileSpec);
				fileCreated := (err = noErr);
			end; (* if *)
			
			if err = noErr then begin
				err := SaveComponentToFile(extensionFileSpec);
			end; (* if *)
			
			if err = noErr then begin
				extensionRefNum := HOpenResFile(extensionFileSpec.vRefNum,
																		extensionFileSpec.parID, 
																			extensionFileSpec.name, fsRdPerm);
				err := ResError;
			end; (* if *)
			if err = noErr then begin
				prefCount := Count1Resources('Pref');
				err := ICBegin(gInstance, icReadWritePerm);
				if err = noErr then begin
					for prefResourceIndex := 1 to prefCount do begin
						UseResFile(extensionRefNum);
						prefData := Get1IndResource('Pref', prefResourceIndex);
						if prefData <> nil then begin
							GetResInfo(prefData, junkID, junkType, prefName);
							DetachResource(prefData);
							junk := ICSetPrefHandle(gInstance, prefName, ICattr_no_change, prefData);
							DisposeHandle(prefData);
						end; (* if *)
					end; (* for *)
					junk := ICEnd(gInstance);
				end; (* if *)

				CloseResFile(extensionRefNum);
			end; (* if *)
			
			if fileCreated then begin
				junk := HDelete(extensionFileSpec.vRefNum, extensionFileSpec.parID, extensionFileSpec.name);
			end; (* if *)
			AddDefaultPrefs := noErr;
		end; (* AddDefaultPrefs *)
		
		var
			err: OSStatus;
			junkInstance : Ptr;
	begin
		err := DoCloseDocument;
		if err = noErr then begin
			(* Reset the window positions to the default, so that the new
				document doesn't inherit the old documents window positions.
			*)
			WindowsResetPositions;
			gDirtyDocument := false;
			gLockedDocument := false;
			err := CreateScratchFile(nil);
			if err = noErr then begin
				if ICGetComponentInstance(gInstance, junkInstance) <> noErr then begin
					err := AddDefaultPrefs;
				end; (* if *)
			end; (* if *)
			if err = noErr then begin
				err := WindowOpen(WT_Main);
			end; (* if *)
			if err = noErr then begin
				WindowSetTitle(WT_Main, GetIndStr(rMiscStrings, strUntitled));
				(* Clear the current file name, thereby setting IsDocumentNew, which
					signifies to the rest of the system that this document is untitled.
				*)
				gCurrentFile.name := '';
			end; (* if *)
			CurrentSeed(gSeedForCurrentFile);
		end; (* if *)
		DoNewDocument := err;
	end; (* DoNewDocument *)

	function DoOpenDocument (fss: FSSpec): OSStatus;
		(* See comment in interface part. *)
		var
			err: OSStatus;
	begin
		err := DoCloseDocument;
		if err = noErr then begin
			gDirtyDocument := false;
			gLockedDocument := FileLocked(fss);
			err := CreateScratchFile(@fss);
			if err = noErr then begin
				WindowsRestorePositions;
				err := WindowOpen(WT_Main);
			end; (* if *)
			if err = noErr then begin
				WindowSetTitle(WT_Main, fss.name);
				gCurrentFile := fss;
			end; (* if *)
			CurrentSeed(gSeedForCurrentFile);
		end; (* if *)
		DoOpenDocument := err;
	end; (* DoOpenDocument *)

	function DoOpenDocumentUsingStandardFile: OSStatus;
		(* See comment in interface part. *)
		var
			err: OSStatus;
			fss: FSSpec;
			junkInfo: FInfo;
	begin
		err := ICStandardGetFile(ICfiletype, fss, junkInfo);
		if err = noErr then begin
			err := DoOpenDocument(fss);
		end; (* if *)
		DoOpenDocumentUsingStandardFile := err;
	end; (* DoOpenDocumentUsingStandardFile *)

	(* ***** Closing and Saving ***** *)

	function DoCloseDocumentWindow (wp: WindowPtr): OSStatus;
		(* See comment in interface part. *)
		var
			wt: WindowType;
			err: OSStatus;
	begin
		wt := MapWindowPtrToWindowType(wp);
		if wt = WT_Main then begin
			err := DoCloseDocument;
		end else begin
			err := WindowClose(wt);
		end; (* case *)
		DoCloseDocumentWindow := err;
	end; (* DoCloseDocumentWindow *)
	
	function DoCloseDocument: OSStatus;
		(* See comment in interface part. *)
		var
			hitItem: integer;
			err: OSStatus;
			junk: OSStatus;
	begin
		(* Closes all the information windows but not WT_Main or WT_About. *)
		err := WindowsCloseAll;
		if (err = noErr) & IsDocumentDirty then begin
			ParamText(GetDocumentName, '', '', '');
			InitCursor;
			hitItem := CautionAlert(rSaveChangesAlert, gOKCancelDiscardModalFilter);
			case hitItem of
				ditOK:
					begin
						err := DoSaveDocument;
					end;
				ditCancel: 
					err := userCanceledErr;
				otherwise
			end; (* case *)
		end; (* if *)
		if IsDocumentOpen then begin
			if err = noErr then begin
				err := WindowClose(WT_Main);
			end; (* if *)
			if err = noErr then begin
				junk := HDelete(gScratchFile.vRefNum, gScratchFile.parID, gScratchFile.name);
				gScratchFile.name := '';
			end; (* if *)
		end; (* if *)
		DoCloseDocument := err;
	end; (* DoCloseDocument *)

	function DoQuit : OSStatus;
		(* See comment in interface part. *)
		var
			err : OSStatus;
	begin
		err := DoCloseDocument;
		if err = noErr then begin
			gQuitNow := true;
		end; (* if *)
		DoQuit := err;
	end; (* DoQuit *)
	
	(* ***** Saving Utilities ***** *)

	function CopyFileSafe (source : FSSpec; dest: FSSpec): OSStatus;
		(* InternalDoSave uses this routine to actually save the scratch
			file back on top of the current document file. This is done
			in the safest possible way, ie by copying the scratch file
			to a temporary file in the same folder as the current file,
			then by deleting the current file and renaming the scratch
			file.
			
			We could have used PBExchangeFiles, but this is a System
			7 feature, so we would have had to implement this code anyway.
			Besides, PBExchangeFiles has a habbit of not working
			as designed on non-standard file systems, eg AUX, Windows NT's
			AFP server, and so on.
		*)
		var
			err : OSStatus;
			junk: OSStatus;
			tempFile: FSSpec;
	begin
		(* Create the temporary file. *)
		err := CreateTemporaryFile(dest.vRefNum, dest.parID, tempFile);
		
		if err = noErr then begin
			(* Copy the scratch file to the temporary. *)
			err := CopyFile(source, tempFile);
			if err = noErr then begin
				(* Remove the dest file and rename the temp to its name. *)
				junk := HDelete(dest.vRefNum, dest.parID, dest.name);
				err := HRename(dest.vRefNum, dest.parID, tempFile.name, dest.name);
			end; (* if *)
		end; (* if *)
		CopyFileSafe := err;
	end; (* CopyFileSafe *)

	function InternalDoSave (fss: FSSpec): OSStatus;
		(* This routine is the core implementation of the saving code. It doesn't
			ask any questions, it just saves the current document right on
			top of the file spec given.
		*)
		var
			err: OSStatus;
			junk : OSStatus;
	begin
		(* WindowsFlushAll calls all the windows WhatFlush routines to push any
			data held in memory out to the ICInstance that points to the scratch file.
		*)
		err := WindowsFlushAll;
		if err = noErr then begin
			(* Now push out the window positions. *)
			WindowsSavePositions;
			(* Hack to force IC to push out any cached information before we
				copy the file.  Basically IC will cache information about the file
				until we either kill the instance or switch the instance to
				a different file.  This is not good, because we're about to
				copy the file using the file system (ie behind IC's back).  The
				workaround is to switch the instance to another file and
				then switch it back immediately.  This is, of course, a hack.
				
				The real solution is to get the IC application to not use
				the file system for the safe save mechanism.  I have a plan
				for this, but I can't do it in this timeframe (IC 1.4).
			*)
			junk := ICSpecifyConfigFile(gInstance, fss);
			junk := ICSpecifyConfigFile(gInstance, gScratchFile);
			(* Now copy the scratch file back on top of the destination file. *)
			err := CopyFileSafe(gScratchFile, fss);
		end; (* if *)
		(* Finally clean up with some bookkeeping stuff. *)
		if err = noErr then begin
			WindowSetTitle(WT_Main, fss.name);
			gDirtyDocument := false;
			gCurrentFile := fss;
		end; (* if *)
		CurrentSeed(gSeedForCurrentFile);
		InternalDoSave := err;
	end; (* InternalDoSave *)

	(* ***** Saving ***** *)

	function DoSaveDocument: OSStatus;
		(* See comment in interface part. *)
		var
			err: OSStatus;
			seedValueNow: longint;
			hitItem : integer;
	begin
		if gCurrentFile.name = '' then begin
			(* New files go through the Save As path. *) 
			err := DoSaveAsDocument;
		end else begin
			(* If the seed has changed, ask the user whether they want to overwrite
				any possible changes.
			*)
			CurrentSeed(seedValueNow);
			if SameSeed(seedValueNow, gSeedForCurrentFile) then begin
				hitItem := ditOK;
			end else begin
				hitItem := CautionAlert(rWarnChangedAlert, gOKCancelModalFilter);
			end; (* if *)
			(* If all is OK, save the file. *)
			if hitItem = ditOK then begin
				err := InternalDoSave(gCurrentFile);
			end; (* if *)
		end; (* if *)
		DoSaveDocument := err;
	end; (* DoSaveDocument *)

	function DoSaveAsDocument: OSStatus;
		(* See comment in interface part. *)
		var
			err: OSStatus;
			fss: FSSpec;
	begin
		err := ICStandardPutFile('', GetDocumentName, fss);
		if err = noErr then begin
			err := InternalDoSave(fss);
		end; (* if *)
		DoSaveAsDocument := err;
	end; (* DoSaveAsDocument *)

	(* ***** Misc Stuff ***** *)

	function CreateSpecifiedInstance(var inst : ICInstance) : OSStatus;
		(* See comment in interface part. *)
		var
			err : OSStatus;
			junk : OSStatus;
	begin
		err := noErr;
		if IsDocumentOpen then begin
			inst := gInstance;
		end else begin
			(* We have to create a temporary instance that points at the default
				config file.
			*)
			inst := nil;
			err := ICStart(inst, ICcreator);
			if err = noErr then begin
				err := ICFindConfigFile(inst, 0, nil);
			end; (* if *)
			
			(* Clean up if the above attempt failed. *)
			if (err <> noErr) and (inst <> nil) then begin
				junk := ICStop(inst);
				inst := nil;
			end; (* if *)
		end; (* if *)
		CreateSpecifiedInstance := err;
	end; (* CreateSpecifiedInstance *)
	
	procedure DestroySpecifiedInstance(inst : ICInstance);
		(* See comment in interface part. *)
		var
			junk : OSStatus;
	begin
		if inst <> gInstance then begin
			junk := ICStop(inst);
		end; (* if *)
	end; (* DestroySpecifiedInstance *)
	
	function InitICDocument: OSStatus;
		(* See comment in interface part. *)
		var
			inst: ICInstance;
			err: OSStatus;
	begin
		gInstance := nil;
		gCurrentFile.name := '';
		gScratchFile.name := '';
		err := ICStart(inst, ICcreator);
		if err = noErr then begin
			gInstance := inst;
		end; (* if *)
		InitICDocument := err;
	end; (* InitICDocument *)

	procedure TermICDocument;
		(* See comment in interface part. *)
		var
			junk: ICError;
	begin
		if gInstance <> nil then begin
			junk := ICStop(gInstance);
		end; (* if *)
	end; (* TermICDocument *)

end. (* ICDocument *)