unit ICMovableModal;

(*	This file is part of the Internet Configuration system and is placed in the public domain for the benefit of all.

	This module implements a MovableModalDialog routine, similar to the Toolbox
	ModalDialog routine in almost all respects, except that it puts up a movable
	modal dialog.  Cool huh?
	
	Loosely based on code by Merzwaren <piovanel@dsi.unimi.it>, since moved into
	Peter's libraries, then into Internet Config.
*)

interface

	uses
		Types,
		Dialogs;

	procedure DisableMenuBar (var movableModalSavedMenuBarState: univ Ptr);
		(* You must call this routine after creating the dialog but before before
			calling MovableModalDialog.  It disables the menu bar, leaving only the
			Edit menu active.
		*)
	
	procedure MovableModalDialog (modalFilter: ModalFilterProcPtr; var itemHit: Integer);
		(* This routine works in an almost identical fashion to the Toolbox routine
			ModalDialog.  One big difference is that you pass in a ModalFilterProcPtr,
			not a UPP.  Because this routine is implemented inside IC, you don't
			need to get involved with MixedMode when calling it.
		*)

	procedure ReEnableMenuBar (var movableModalSavedMenuBarState: univ Ptr);
		(* You must call this routine after finishing up with the dialog but
			before you dispose it.  This routine resets the menu bar back to the saved
			state. 
		*)

implementation

	uses
		LowMem,
		Dialogs, 
		ToolUtils, 
		Scrap, 

		ICResources,
		ICMiscSubs,
		ICDialogs, 
		ICGlobals;

	(* ***** Enable/Disable Menu Bar Code ***** *)

	const
		kSystemMenuThreshold = -16000;		
		(* Menu IDs <= than kSystemMenuThreshold are used by the system and we
			don't mess around with them.
		*)
		kMovableModalEventMask = mDownMask + mUpMask + 
													keyDownMask + keyUpMask + autoKeyMask + 
													updateMask + activMask + osMask;
		(* kMovableModalEventMask is a mask for the event that we handle in
			our MovableModalDialog routine.
		*)

	(* MenuListHandle is the type of the system menu list data structure, ie
		the thing contained in the MenuList low memory global.  Obviously
		such a data structure must be aligned using 68K conventions.
	*)

{$PUSH}
{$ALIGN MAC68K}
	
	type
		MenuEntry =
			record
				hMenu: MenuHandle;
				leftEdge: Integer;
			end;
		MenuList =
			record
				offsetToLastMenu: Integer;
				unused1: Integer;
				unused2: Integer;
				theMenus: array[0..1000] of MenuEntry;
			end;
		MenuListPtr = ^MenuList;
		MenuListHandle = ^MenuListPtr;

{$ALIGN RESET}
{$POP}

	(* MenuBarState is the internal description of what the external opaque pointer,
		movableModalSavedMenuBarState, points to.  In DisableMenuBar we record a set
		of bits (mbsBarEnable) that determine which menus are enabled/disabled before
		disabling them.  In EnableMenuBar we reenable the appropriate menus.  We
		also do the same thing for the items on the Edit menu, storing their state
		in mbsEditEnable.
	*)
	type
		MenuBarState =
			record
				mbsBarEnable: LongInt;
				mbsEditEnable: LongInt;
			end;
		MenuBarStatePtr = ^MenuBarState;

	procedure DisableMenuBar (var movableModalSavedMenuBarState: univ Ptr);
		(* See comment in interface part. *)
		var
			state: MenuBarStatePtr;
			theDialog: DialogPtr;
			needEditMenu: Boolean;
			menuList: MenuListHandle;
			numberOfMenus: Integer;
			indexIntoMenuList : integer;
			theMenu: MenuHandle;
			menuID: Integer;
	begin
		(* Setup a bunch of variables. *)
		theDialog := FrontWindow;
		needEditMenu := (theDialog <> nil) & (GetSelectedDialogTextItem(theDialog) > 0);
		menuList := MenuListHandle(LMGetMenuList);
		numberOfMenus := menuList^^.offsetToLastMenu div sizeof(MenuEntry);

		(* Now allocate the memory for the MenuBarState. *)
		movableModalSavedMenuBarState := NewPtr(sizeof(MenuBarState));
		state := MenuBarStatePtr(movableModalSavedMenuBarState);

		(* Walk the menu list disabling all the enabled menus and setting the
			the bits in the MenuBarState to record what damage we've done.
		*)
		state^.mbsBarEnable := 0;
		for indexIntoMenuList := 0 to numberOfMenus - 1 do begin
			theMenu := menuList^^.theMenus[indexIntoMenuList].hMenu;
			menuID := theMenu^^.menuID;

			if (menuID > kSystemMenuThreshold) then begin
				(* Only do this stuff if we're not a system menu. *)

				if menuID = mEdit then begin
					(* Take out the edit menu as a special case.  We have to record the
						state of all the items on the edit menu and then enable the
						menu items we're prepared to handle.
					*)
					state^.mbsEditEnable := theMenu^^.enableFlags;
					if needEditMenu then begin
						theMenu^^.enableFlags := 1 + BSL(1, iCut) + BSL(1, iCopy) + BSL(1, iPaste);
					end else begin
						DisableItem(theMenu, 0);
					end; (* if *)
				end else begin
					(* Make sure the menu item is disabled and adjust the balloon help appropriately. *)
					if (btst(theMenu^^.enableFlags, 0)) then begin
						(* If this menu is enabled, disable it and set the corresponding bit. *)
						bset(state^.mbsBarEnable, indexIntoMenuList);
						DisableItem(theMenu, 0);
					end; (* if *)
				end; (* if *)
			end; (* if *)
		end; (* for *)

		(* Clean up. *)
		HiliteMenu(0);
		DrawMenuBar;
	end; (* DisableMenuBar *)

	procedure ReEnableMenuBar (var movableModalSavedMenuBarState: univ Ptr);
		(* See comment in interface part. *)
		var
			state: MenuBarStatePtr;
			menuList: MenuListHandle;
			numberOfMenus: Integer;
			indexIntoMenuList : integer;
			theMenu: MenuHandle;
			menuID: Integer;
	begin
		(* Set up some local variables. *)
		state := MenuBarStatePtr(movableModalSavedMenuBarState);
		menuList := MenuListHandle(LMGetMenuList);
		numberOfMenus := menuList^^.offsetToLastMenu div sizeof(MenuEntry);

		(* Index through each of the menus in the menu list
			re-enabling them if appropriate.
		*)
		for indexIntoMenuList := 0 to numberOfMenus - 1 do begin
			theMenu := menuList^^.theMenus[indexIntoMenuList].hMenu;
			menuID := theMenu^^.menuID;

			if menuID > kSystemMenuThreshold then begin
				(* Only do this stuff if we're not a system menu. *)
				if menuID = mEdit then begin
					theMenu^^.enableFlags := state^.mbsEditEnable;
				end else if btst(state^.mbsBarEnable, indexIntoMenuList) then begin
					EnableItem(theMenu, 0);
				end; (* if *)
			end; (* if *)
		end; (* for *)

		(* Clean up. *)
		DisposePtr(movableModalSavedMenuBarState);
		movableModalSavedMenuBarState := nil;
		DrawMenuBar;
	end; (* ReEnableMenuBar*)

	(* ***** Movable Modal Dialog Code ***** *)

	function DoMenuChoice (menuAndItem: LongInt; theDialog: DialogPtr; var itemHit: Integer): Boolean;
		(* This routine is called in response to a menu event in the dialog. The routine
			implements the standard Edit menu functionality in the dialog.  The meaning
			of theDialog, itemHit and return result and vaguely similar to that of a filter
			function, ie it returns true if it consumes the event and sets itemHit to
			the dialog item that consumed it.
		*)
		var
			result : Boolean;
			currentEditField : integer;
			itemKind: integer;
			junk: OSStatus;
	begin
		result := false;
		(* System 7 handles menu operations in dialogs for us,
			so we have to explicitly disable our code if we're running
			under 7.
		*)
		if (hiwrd(menuAndItem) = mEdit) & not gHasSystem7 then begin
			currentEditField := GetSelectedDialogTextItem(theDialog);
			GetDItemKind(theDialog, currentEditField, itemKind);

			if (band(itemKind, itemDisable) = 0) then begin
 				(* if the current edit field is an enabled item, 
 					let MovableModalDialog know we wish to exit.
 				*)
 				result := true;
				itemHit := currentEditField;
			end; (* if *)

			case lowrd(menuAndItem) of
				iCut: 
					begin
						DialogCut(theDialog);
						junk := ZeroScrap;
						junk := TEToScrap;
					end;
				iCopy: 
					begin
						DialogCopy(theDialog);
						junk := ZeroScrap;
						junk := TEToScrap;
					end;
				iPaste: 
					begin
						junk := TEFromScrap;
						DialogPaste(theDialog);
					end;
				otherwise
					(* do nothing *) ;
			end; (* case *)
		end; (* if *)

		HiliteMenu(0);
		DoMenuChoice := result;
	end; (* DoMenuChoice *)

	function HandleMouseDown (theDialog: DialogPtr; var theEvent: EventRecord; var itemHit: Integer): Boolean;
		(* This routine is called in response to a mouseDown event in the movable modal
			dialog.  Most of the real work to handle mouse downs has already been done
			by DialogSelect (which is called after this routine).  All we have to do
			is handle those extra cases that DialogSelect doesn't.
		*)
		var
			result : Boolean;
			partCode: Integer;
			foundWindow: WindowPtr;
			dialogBeeperProc: UniversalProcPtr;
			dragRect: Rect;
	begin
		result := false;
		partCode := FindWindow(theEvent.where, foundWindow);

		if partCode = inMenuBar then begin
			(* Handle menu clicks. *)
			result := DoMenuChoice(MenuSelect(theEvent.where), theDialog, itemHit);
		end else if not PtInRgn(theEvent.where, WindowPeek(theDialog)^.strucRgn) then begin
			(* Handle clicks outside of the window. *)
			dialogBeeperProc := LMGetDABeeper;
			if dialogBeeperProc <> nil then begin
				CallSoundProc(1, dialogBeeperProc);
			end; (* if *)
		end else if (partCode = inDrag) & (theDialog = foundWindow) then begin
			(* Now we have to Handle the only thing DialogSelect doesn't do for us: dragging. *)
			dragRect := GetGrayRgn^^.rgnBBox;
			DragWindow(foundWindow, theEvent.where, dragRect);
			theEvent.what := nullEvent;
		end; (* if *)
		HandleMouseDown := result;
	end; (* HandleMouseDown *)

	procedure MovableModalDialog (modalFilter: ModalFilterProcPtr; var itemHit: Integer);
		(* See comment in interface part. *)
		var
			theDialog : DialogPtr;
			junkDialog: DialogPtr;
			theEvent: EventRecord;
			junkBool: Boolean;
			timeToLeave : Boolean;
	begin
		itemHit := 0;
		theDialog := FrontWindow;
		if theDialog <> nil then begin
			SetPort(theDialog);
			repeat
				junkBool := WaitNextEvent(kMovableModalEventMask, theEvent, 0, nil);
				SetPort(theDialog); (* Just in case some loser DA has switch it on us! *)

				if (modalFilter <> nil) & modalFilter(theDialog, theEvent, itemHit) then begin
					timeToLeave := true;
				end else if (theEvent.what = mouseDown) & HandleMouseDown(theDialog, theEvent, itemHit) then begin
					timeToLeave := true;
				end else if (theEvent.what = keyDown) & (band(theEvent.modifiers, cmdKey) <> 0) 
						& DoMenuChoice(MenuKey(chr(band(theEvent.message, charCodeMask))), theDialog, itemHit) then begin
					timeToLeave := true;
				end else if IsDialogEvent(theEvent) & DialogSelect(theEvent, junkDialog, itemHit) then begin
					timeToLeave := true;
				end; (* if *)
				
			until timeToLeave;
		end; (* if *)
	end; (* MovableModalDialog *)

end. (* ICMovableModal *)
