unit ICStandardFile;

(*	This file is part of the Internet Configuration system and is placed in the public domain for the benefit of all.

	This routine provides a set of wrappers around the system Standard File
	routines.  We need these because:
	
	1. There are different system routines for System 6 and System 7.
	2. The system routine's API are kinda weird.
	3. There's no system StandardGetFolder

	All of these routines return userCanceledErr if the user cancels the dialog.
	
	One nasty trick in this module is the Point($00640064) construct.  This
	creates a Point data type where each co-ordinate is 100 decimal, perfect
	for putting the standard file dialog in a place you can see.  We only use
	this for System 6 dialogs.  We presume that System 7 dialogs will look
	after themselves and we really don't care how good System 6 looks.
*)

interface

	uses
		Files,
		
		ICResources,
		ICDialogs;
	
	function ICStandardGetFile (requestedFileType: OSType; 
												var chosenFileSpec: FSSpec;
												var infoForChosenFile: FInfo): OSStatus;
		(* This routine prompts the user for a file of type requestedFileType and returns
			a reference to that file in chosenFileSpec.  It also returns the file's Finder Info,
			because routines that call this routine often need that information.
		*)
	
	function ICStandardPutFile (prompt, name: Str255; var chosenFileSpec: FSSpec): OSStatus;
		(* This routine prompts the user for a destination for a file to be saved.
			The call provides a prompt string for the dialog and a default name string
			for the file.  The file's final destination is return in chosenFileSpec.
		*)
	
	function ICStandardGetFolder (var chosenFolderSpec: FSSpec;
														var dirIDOfChosenFolder: longint): OSStatus;
		(* This routine prompts the user to choose a folder.  It returns
			a reference to the chosen folder.  It also returns the dirID of the chosen
			folder (as opposed to the parID in the FSSpec, which is the dirID of
			the folder that contains the chosen folder).  The jury is still up in the
			air as to whether this is needed by ICStandardGetFolder single
			client (the FSSpecWhat), but until I decide that issue, the parameter
			lives.
		*)

	procedure InitICStandardFile;
		(* Initialises the module. *)
		
implementation

	uses
		Finder,
		StandardFile, 
		LowMem,
		TextUtils,
		Aliases,
		
		ICGlobals, 
		ICMiscSubs;

	var
		gStandardGetFolderFileFilterUPP : FileFilterYDUPP;
		gStandardGetFolderDialogUPP : DlgHookYDUPP;
		gSFPGetFileButtonUPP : DlgHookUPP;
		
	(* ***** StandardGetFolder for System 7 ****** *)

	(* Original comment:
	
		This little chunk o' code implements a way to let the user choose a       
		folder to save files in via a StandardFile Dialog.

		Since the code uses the CustomGetFile function and depends on the use of  
		FSSpec records, it only works under System 7.0 or later.                  
  
		Portions of this code were originally provided by Paul Forrester
		<paulf@apple.com> to the think-c Internet mailing list in response to my
		my question on how to do exactly what this code does.  I've added a
		couple of features, such as the ability to Handle aliased folders and
		the programmer definable prompt.  I also cleaned and tightened the code,
		stomped a couple of bugs, and packaged it up neatly.  Bunches of work,
		but I learned A LOT about Standard File, the File Manager, the Dialog
		Manager, and the Alias Manager.  I tried to include in the comments some
		of the neat stuff I discovered in my hours of pouring over Inside Mac.
		Hope you find it educational as well as useful.

		Converted to Pascal by Peter N Lewis <peter.lewis@info.curtin.edu.au>, Dec 1992.
		
		Cleaned up and integrated into IC by Quinn <quinn@quinn.echidna.id.au>,  Oct 1996
	*)

	type
		StandardFileReplyPtr = ^StandardFileReply;

	var
		gCurrentSelectedFolder: Str255;
		gStandardGetFolderPrompt: Str255;

	function StandardGetFolderFileFilter (cpbPtr: CInfoPBPtr; ignored: Ptr): Boolean;
		(* This is the file filter passed to CustomGetFile. It passes folders only. *)
	begin
		{$unused ignored}
		StandardGetFolderFileFilter := (band(cpbPtr^.ioFlAttrib, ioDirMask) = 0) | 
															(band(cpbPtr^.ioDrUsrWds.frFlags, $4000) <> 0);
	end; (* StandardGetFolderFileFilter *)

	function StandardGetFolderDialogHook (hitItem: integer; dlg: DialogPtr; myDataPtr: StandardFileReplyPtr) : integer;
		(* This function lets us process item hits in the GetFolderDialog.  We're
			only interested if the user hit the ditGetFolderSelectButton. We pass all
			other item hits back unmodified.
		*)

		procedure SetButtonTitle (newName: Str255);
			(* This function sets the title of the "Select Wibble" button to newName
				truncated appropriately.  It also updates the gCurrentSelectedFolder
				with the full name, so that we can check whether we need to
				do this work the next time around.
			*)
			var
				buttonNameProto : Str255;
				placeHolderIndex: integer;
				leftPart, rightPart: Str255;
				itemRect: Rect;
				junkResultCode: integer;
				widthRemaining: integer;
		begin
			if gCurrentSelectedFolder <> newName then begin
				gCurrentSelectedFolder := newName;

				(* Get the prototype for the button name.  This is stored in the
					text of an offscreen static text item in the dialog. Then
					pull it apart by searching for ^1 and copying out the left and
					right parts.
				*)
				GetDItemText(dlg, ditGetFolderSelectString, buttonNameProto);
				
				placeHolderIndex := pos('^1', buttonNameProto);
				leftPart := TPCopy(buttonNameProto, 1, placeHolderIndex - 1);
				rightPart := TPCopy(buttonNameProto, placeHolderIndex + 2, 255);
				
			    (* Find the width left over in the button after drawing the word "Select"
			    	the quotation marks. Truncate the new name to this width.
			    *)
				GetDItemRect(dlg, ditGetFolderSelectButton, itemRect);
				widthRemaining := (itemRect.right - itemRect.left) - StringWidth(concat(' ', leftPart, rightPart, ' '));
				junkResultCode := TruncString(widthRemaining, newName, smTruncEnd);

				(* Finally set the buttons title. *)
				SetDControlTitle(dlg, ditGetFolderSelectButton, concat(leftPart, newName, rightPart));
				ValidRect(itemRect);
			end; (* if *)
		end; (* SetButtonTitle *)

		procedure SetFolderButtonTitle (vRefNum: integer; dirID: longint);
			(* This function sets the name of the "Select wibble" button
				based on the vRefNum and and dirID of a folder.
			*)
			var
				err: OSStatus;
				cpb: CInfoPBRec;
				folderName: Str63;
		begin
			cpb.ioFDirIndex := -1;
			cpb.ioNamePtr := @folderName;
			cpb.ioVRefNum := vRefNum;
			cpb.ioDirID := dirID;
			err := PBGetCatInfoSync(@cpb);
			if err = noErr then begin
				SetButtonTitle(folderName);
			end; (* if *)
		end; (* SetFolderButtonTitle *)

	begin
	    (* CustomGetFile calls dialog hook for both main and subsidiary dialog boxes.
			Make sure that dialog record indicates that this is the main GetFolder dialog.
		*)
		if OSType(WindowPeek(dlg)^.refCon) = sfMainDialogRefCon then begin

			if hitItem = sfHookFirstCall then begin
				(* This is the first call, so setup the prompt above the file list. *)
				SetDItemText(dlg, ditGetFolderPromptString, gStandardGetFolderPrompt);
			end else begin
				(* This is a subsequent call. *)
				if myDataPtr^.sfFile.name = '' then begin
					(* If the current name is empty, the following fields aren't always set properly
						so get them out of low memory.
					*)
					myDataPtr^.sfFile.vRefNum := - LMGetSFSaveDisk;
					myDataPtr^.sfFile.parID := LMGetCurDirStore;
					myDataPtr^.sfFile.name := '';
					SetFolderButtonTitle(myDataPtr^.sfFile.vRefNum, myDataPtr^.sfFile.parID);
				end else begin
					SetButtonTitle(myDataPtr^.sfFile.name);
				end; (* if *)
			end; (* if *)

			(* If they click on the "Select Wibble" button, tell Standard File to bail
				out by returning the Cancel button, but set the sfGood Boolean so
				that we know to use the results.
			*)
			if hitItem = ditGetFolderSelectButton then begin
				hitItem := sfItemCancelButton;
				myDataPtr^.sfGood := true;
			end; (* if *)

		end; (* if *)
		StandardGetFolderDialogHook := hitItem;
	end; (* StandardGetFolderDialogHook *)

	procedure StandardGetFolder (prompt : Str255; var mySFReply: StandardFileReply);
		(* The StandardGetFolder function. You pass it the prompt to display above the file
			list, and a place to put a StandardFileReply record.

			Upon return, the sfFile field of the SFReply record contains the file
			spec of the chosen folder.  The sfGood field is set to true if the user
			chose a folder, or false if not.
		*)
		var
			err: OSStatus;
			dummyTypeList: SFTypeList;
			junkBool1, junkBool2: Boolean;
			cpb: CInfoPBRec;
	begin
		gStandardGetFolderPrompt := prompt;
		gCurrentSelectedFolder := '';
		CustomGetFile(gStandardGetFolderFileFilterUPP,	(* filters out everything except folders *)
								-1, @dummyTypeList, 						(* dummy type list *)
								mySFReply, 										(* space for result *)
								rCustomGetFolderDialog, 					(* dialog resource *)
								Point($00640064),							(* place to put window, system overrides this *)
								gStandardGetFolderDialogUPP, 		(* dialog hook drives our bonus items *)
								nil, 													(* no modal filter proc *)
								nil, 													(* no activation order list *)
								nil, 													(* no activate proc *)
								@mySFReply);									(* our data pointer *)

	    (* Now the reply record contains the volume reference number and the
			name of the selected folder. We need to use PBGetCatInfo to get the
			directory ID of the selected folder.
		*)
		if mySFReply.sfGood then begin
			
			(* If we got a name-based response (ie the user clicked the Open button
				rather than the Select Wibble button), make sure to resolve any aliases
				before continuing.
			*)
			if mySFReply.sfFile.name <> '' then begin
				err := ResolveAliasFile(mySFReply.sfFile, true, junkBool1, junkBool2);
				if err = noErr then begin
					(* Now look up the dirID of the folder we're pointed at. *)
					cpb.ioVRefNum := mySFReply.sfFile.vRefNum;
					cpb.ioDirID := mySFReply.sfFile.parID;
					cpb.ioNamePtr := @mySFReply.sfFile.name;
					cpb.ioFDirIndex := 0;
					err := PBGetCatInfoSync(@cpb);
				end; (* if *)

				mySFReply.sfGood := (err = noErr);
				mySFReply.sfFile.parID := cpb.ioDrDirID;
				mySFReply.sfFile.name := '';
			end; (* if *)
			
			(* Now call GetCatInfo to find the real FSSpec for the vRefNum/dirID pair we
				currently have.
			*)
			if (err = noErr) and mySFReply.sfGood then begin
				cpb.ioVRefNum := mySFReply.sfFile.vRefNum;
				cpb.ioDirID := mySFReply.sfFile.parID;
				cpb.ioNamePtr := @mySFReply.sfFile.name;
				cpb.ioFDirIndex := -1;
				err := PBGetCatInfoSync(@cpb);
			end; (* if *)
		end; (* if *)
	end; (* StandardGetFolder *)

	(* ***** API Routines ***** *)

	function ICStandardGetFile (requestedFileType: OSType; 
												var chosenFileSpec: FSSpec;
												var infoForChosenFile: FInfo): OSStatus;
		(* See comment in interface part. *)
		var
			err: OSStatus;
			typeList: SFTypeList;
			numberOfTypes: integer;
			oldReply: SFReply;
			newReply: StandardFileReply;
			junkLong: longint;
	begin
		(* Calculate typeList and numberOfTypes based on input parameters. *)
		typeList[0] := requestedFileType;
		if requestedFileType = OSType(0) then begin
			numberOfTypes := -1;
		end else if requestedFileType = 'APPL' then begin
			typeList[1] := kApplicationAliasType;
			numberOfTypes := 2;
		end else begin
			numberOfTypes := 1;
		end; (* if *)
		
		(* Now call either the new or old standard file routines. *)
		if gHasNewStandardFile then begin
			StandardGetFile(nil, numberOfTypes, @typeList, newReply);
			if newReply.sfGood then begin
				chosenFileSpec := newReply.sfFile;
				err := HGetFInfo(chosenFileSpec.vRefNum, chosenFileSpec.parID, 
											chosenFileSpec.name, infoForChosenFile);
			end else begin
				err := userCanceledErr;
			end; (* if *)
		end else begin
			SFGetFile(Point($00640064), '', nil, numberOfTypes, @typeList, nil, oldReply);
			if oldReply.good then begin
				err := GetFInfo(oldReply.fName, oldReply.vRefNum, infoForChosenFile);
				if err = noErr then begin
					chosenFileSpec.name := oldReply.fName;
					err := GetWDInfo(oldReply.vRefNum, chosenFileSpec.vRefNum, chosenFileSpec.parID, junkLong);
				end; (* if *)
			end else begin
				err := userCanceledErr;
			end; (* if *)
		end; (* if *)
		
		ICStandardGetFile := err;
	end; (* ICStandardGetFile *)

	function ICStandardPutFile (prompt, name: Str255; var chosenFileSpec: FSSpec): OSStatus;
		(* See comment in interface part. *)
		var
			err: OSStatus;
			oldReply: SFReply;
			newReply: StandardFileReply;
			junkLong: longint;
	begin
		(* Call either the new or old standard file routines. *)
		if gHasNewStandardFile then begin
			StandardPutFile(prompt, name, newReply);
			if newReply.sfGood then begin
				chosenFileSpec := newReply.sfFile;
				err := noErr;
			end else begin
				err := userCanceledErr;
			end; (* if *)
		end else begin
			SFPutFile(Point($00640064), prompt, name, nil, oldReply);
			if oldReply.good then begin
				chosenFileSpec.name := oldReply.fName;
				err := GetWDInfo(oldReply.vRefNum,
											chosenFileSpec.vRefNum, chosenFileSpec.parID, junkLong);
			end else begin
				err := userCanceledErr;
			end; (* if *)
		end; (* if *)
		ICStandardPutFile := err;
	end; (* ICStandardPutFile *)

	function SFPGetFileButtonHook (item: integer; dlg: DialogPtr): integer;
		(* We pass this routine to SFPGetFile to operate the This Folder
			button.  It simply translates hits on that button to hits on
			the Open button.  This causes the dialog to exit, and we then
			use the item that's in the SFReply.
		*)
	begin
		{$unused dlg}
		if item = ditThisFolder then begin
			item := sfItemOpenButton;
		end; (* if *)
		SFPGetFileButtonHook := item;
	end; (* SFPGetFileButtonHook*)

	function ICStandardGetFolder (var chosenFolderSpec: FSSpec;
														var dirIDOfChosenFolder: longint): OSStatus;
		(* See comment in interface part. *)
		var
			err: OSStatus;
			typeList: SFTypeList;
			oldReply: SFReply;
			newReply: StandardFileReply;
			junkLong: longint;
			cpb: CInfoPBRec;
	begin
		(* Call old or new routines as appropriate. *)
		if gHasNewStandardFile then begin

			(* Under System 7, we have a huge routine that implements this, so we
				just call it.
			*)
			StandardGetFolder('', newReply);
			if newReply.sfGood then begin
				chosenFolderSpec := newReply.sfFile;
				err := noErr;
			end else begin
				err := userCanceledErr;
			end; (* if *)
		end else begin
			
			(* Under System 6, we do a kludge.  Basically we implement a StandardGetFolder
				routine the old fashion way, ie not conforming to the current HI spec.
				Instead we just put up a SFPGetFile dialog with a bonus button that lets
				you select a folder.  We pass SFPGetFileButtonHook into the routine to make the 
				button functional.

 				Note that we can't set numTypes = 0 because it doesn't seem to stop files
 				from being displayed in the list.  Instead we pass in a typeList containing
 				one entry, set to OSType(0).  Presumably we won't have any OSType(0)
 				typed files on the disk (-:
			*)
 			typeList[0] := OSType(0);
			SFPGetFile(Point($00640064), 			(* where = 100,100 *)
								'', 										(* no prompt, it's ignored anyway *)
								nil, 									(* no file filter function *)
								1, @typeList, 					(* see note above *)
								gSFPGetFileButtonUPP, 	(* dialog hook to run the This Folder button *)
								oldReply, 							(* put the chosen file here *)
								rFolderOpenDialog6,			(* resource used for display *)
								nil);									(* no modal filter proc *)
			if oldReply.good then begin
				err := GetWDInfo(oldReply.vRefNum, chosenFolderSpec.vRefNum, chosenFolderSpec.parID, junkLong);
			end else begin
				err := userCanceledErr;
			end; (* if *)
		end; (* if *)
		
		(* Now work out dirIDOfChosenFolder. *)
		if err = noErr then begin
			err := FSpGetCatInfo(chosenFolderSpec, -1, cpb);
			if err = noErr then begin
				chosenFolderSpec.parID := cpb.ioDrParID;
				dirIDOfChosenFolder := cpb.ioDirID;
			end; (* if *)
		end; (* if *)
		
		ICStandardGetFolder := err;
	end; (* ICStandardGetFolder *)

	procedure InitICStandardFile;
	begin
		gStandardGetFolderFileFilterUPP := NewFileFilterYDProc(@StandardGetFolderFileFilter);
		gStandardGetFolderDialogUPP := NewDlgHookYDProc(@StandardGetFolderDialogHook);
		gSFPGetFileButtonUPP := NewDlgHookProc(@SFPGetFileButtonHook);
	end; (* InitICStandardFile *)

end. (* ICStandardFile *)
